@@ -51,6 +51,7 @@ bd_plugins = { "lvm": BlockDev.Plugin.LVM,
                "fs": BlockDev.Plugin.FS,
                "s390": BlockDev.Plugin.S390,
                "nvdimm": BlockDev.Plugin.NVDIMM,
+               "vdo": BlockDev.Plugin.VDO,
 }
 
 
@@ -191,6 +192,28 @@ def btrfs_check(mountpoint, extra=None, **kwargs):
 __all__.append("btrfs_check")
 
 
+class CryptoLUKSPBKDF(BlockDev.CryptoLUKSPBKDF):
+    def __new__(cls, type=None, hash=None, max_memory_kb=0, iterations=0, time_ms=0, parallel_threads=0):  # pylint: disable=redefined-builtin
+        ret = BlockDev.CryptoLUKSPBKDF.new(type, hash, max_memory_kb, iterations, time_ms, parallel_threads)
+        ret.__class__ = cls
+        return ret
+    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
+        super(CryptoLUKSPBKDF, self).__init__()  #pylint: disable=bad-super-call
+CryptoLUKSPBKDF = override(CryptoLUKSPBKDF)
+__all__.append("CryptoLUKSPBKDF")
+
+class CryptoLUKSExtra(BlockDev.CryptoLUKSExtra):
+    def __new__(cls, data_alignment=0, data_device=None, integrity=None, sector_size=0, label=None, subsystem=None, pbkdf=None):
+        if pbkdf is None:
+            pbkdf = CryptoLUKSPBKDF()
+        ret = BlockDev.CryptoLUKSExtra.new(data_alignment, data_device, integrity, sector_size, label, subsystem, pbkdf)
+        ret.__class__ = cls
+        return ret
+    def __init__(self, *args, **kwargs):   # pylint: disable=unused-argument
+        super(CryptoLUKSExtra, self).__init__()  #pylint: disable=bad-super-call
+CryptoLUKSExtra = override(CryptoLUKSExtra)
+__all__.append("CryptoLUKSExtra")
+
 # calling `crypto_luks_format_luks2` with `luks_version` set to
 # `BlockDev.CryptoLUKSVersion.LUKS1` and `extra` to `None` is the same
 # as using the "original" function `crypto_luks_format`
@@ -236,6 +259,22 @@ def crypto_luks_resume(device, passphrase=None, key_file=None):
     return _crypto_luks_resume(device, passphrase, key_file)
 __all__.append("crypto_luks_resume")
 
+_crypto_tc_open = BlockDev.crypto_tc_open_full
+@override(BlockDev.crypto_tc_open)
+def crypto_tc_open(device, name, passphrase, read_only=False, keyfiles=None, hidden=False, system=False, veracrypt=False, veracrypt_pim=0):
+    if isinstance(passphrase, str):
+        passphrase = [ord(c) for c in passphrase]
+    return _crypto_tc_open(device, name, passphrase, keyfiles, hidden, system, veracrypt, veracrypt_pim, read_only)
+__all__.append("crypto_tc_open")
+
+_crypto_bitlk_open = BlockDev.crypto_bitlk_open
+@override(BlockDev.crypto_bitlk_open)
+def crypto_bitlk_open(device, name, passphrase, read_only=False):
+    if isinstance(passphrase, str):
+        passphrase = [ord(c) for c in passphrase]
+    return _crypto_bitlk_open(device, name, passphrase, read_only)
+__all__.append("crypto_bitlk_open")
+
 
 _dm_create_linear = BlockDev.dm_create_linear
 @override(BlockDev.dm_create_linear)
@@ -623,6 +662,61 @@ def lvm_cache_pool_convert(vg_name, data_lv, metadata_lv, name=None, extra=None,
     return _lvm_cache_pool_convert(vg_name, data_lv, metadata_lv, name, extra)
 __all__.append("lvm_cache_pool_convert")
 
+_lvm_vdo_pool_create = BlockDev.lvm_vdo_pool_create
+@override(BlockDev.lvm_vdo_pool_create)
+def lvm_vdo_pool_create(vg_name, lv_name, pool_name, data_size, virtual_size, index_memory=0, compression=True, deduplication=True, write_policy=BlockDev.LVMVDOWritePolicy.AUTO, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _lvm_vdo_pool_create(vg_name, lv_name, pool_name, data_size,virtual_size, index_memory, compression, deduplication, write_policy, extra)
+__all__.append("lvm_vdo_pool_create")
+
+_lvm_vdo_resize = BlockDev.lvm_vdo_resize
+@override(BlockDev.lvm_vdo_resize)
+def lvm_vdo_resize(vg_name, lv_name, size, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _lvm_vdo_resize(vg_name, lv_name, size, extra)
+__all__.append("lvm_vdo_resize")
+
+_lvm_vdo_pool_resize = BlockDev.lvm_vdo_pool_resize
+@override(BlockDev.lvm_vdo_pool_resize)
+def lvm_vdo_pool_resize(vg_name, lv_name, size, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _lvm_vdo_pool_resize(vg_name, lv_name, size, extra)
+__all__.append("lvm_vdo_pool_resize")
+
+_lvm_vdo_enable_compression = BlockDev.lvm_vdo_enable_compression
+@override(BlockDev.lvm_vdo_enable_compression)
+def lvm_vdo_enable_compression(vg_name, pool_name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _lvm_vdo_enable_compression(vg_name, pool_name, extra)
+__all__.append("lvm_vdo_enable_compression")
+
+_lvm_vdo_disable_compression = BlockDev.lvm_vdo_disable_compression
+@override(BlockDev.lvm_vdo_disable_compression)
+def lvm_vdo_disable_compression(vg_name, pool_name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _lvm_vdo_disable_compression(vg_name, pool_name, extra)
+__all__.append("lvm_vdo_disable_compression")
+
+_lvm_vdo_enable_deduplication = BlockDev.lvm_vdo_enable_deduplication
+@override(BlockDev.lvm_vdo_enable_deduplication)
+def lvm_vdo_enable_deduplication(vg_name, pool_name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _lvm_vdo_enable_deduplication(vg_name, pool_name, extra)
+__all__.append("lvm_vdo_enable_deduplication")
+
+_lvm_vdo_disable_deduplication = BlockDev.lvm_vdo_disable_deduplication
+@override(BlockDev.lvm_vdo_disable_deduplication)
+def lvm_vdo_disable_deduplication(vg_name, pool_name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _lvm_vdo_disable_deduplication(vg_name, pool_name, extra)
+__all__.append("lvm_vdo_disable_deduplication")
+
+_lvm_vdo_pool_convert = BlockDev.lvm_vdo_pool_convert
+@override(BlockDev.lvm_vdo_pool_convert)
+def lvm_vdo_pool_convert(vg_name, lv_name, pool_name, virtual_size, index_memory=0, compression=True, deduplication=True, write_policy=BlockDev.LVMVDOWritePolicy.AUTO, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _lvm_vdo_pool_convert(vg_name, lv_name, pool_name, virtual_size, index_memory, compression, deduplication, write_policy, extra)
+__all__.append("lvm_vdo_pool_convert")
 
 _md_get_superblock_size = BlockDev.md_get_superblock_size
 @override(BlockDev.md_get_superblock_size)
@@ -737,6 +831,99 @@ def nvdimm_namespace_disable(namespace, extra=None, **kwargs):
     return _nvdimm_namespace_disable(namespace, extra)
 __all__.append("nvdimm_namespace_disable")
 
+
+_vdo_create = BlockDev.vdo_create
+@override(BlockDev.vdo_create)
+def vdo_create(name, backing_device, logical_size=0, index_memory=0, compression=True, deduplication=True, write_policy=BlockDev.VDOWritePolicy.AUTO, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_create(name, backing_device, logical_size, index_memory, compression, deduplication, write_policy, extra)
+__all__.append("vdo_create")
+
+_vdo_remove = BlockDev.vdo_remove
+@override(BlockDev.vdo_remove)
+def vdo_remove(name, force=False, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_remove(name, force, extra)
+__all__.append("vdo_remove")
+
+_vdo_change_write_policy = BlockDev.vdo_change_write_policy
+@override(BlockDev.vdo_change_write_policy)
+def vdo_change_write_policy(name, write_policy, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_change_write_policy(name, write_policy, extra)
+__all__.append("vdo_change_write_policy")
+
+_vdo_enable_compression = BlockDev.vdo_enable_compression
+@override(BlockDev.vdo_enable_compression)
+def vdo_enable_compression(name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_enable_compression(name, extra)
+__all__.append("vdo_enable_compression")
+
+_vdo_disable_compression = BlockDev.vdo_disable_compression
+@override(BlockDev.vdo_disable_compression)
+def vdo_disable_compression(name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_disable_compression(name, extra)
+__all__.append("vdo_disable_compression")
+
+_vdo_enable_deduplication = BlockDev.vdo_enable_deduplication
+@override(BlockDev.vdo_enable_deduplication)
+def vdo_enable_deduplication(name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_enable_deduplication(name, extra)
+__all__.append("vdo_enable_deduplication")
+
+_vdo_disable_deduplication = BlockDev.vdo_disable_deduplication
+@override(BlockDev.vdo_disable_deduplication)
+def vdo_disable_deduplication(name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_disable_deduplication(name, extra)
+__all__.append("vdo_disable_deduplication")
+
+_vdo_activate = BlockDev.vdo_activate
+@override(BlockDev.vdo_activate)
+def vdo_activate(name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_activate(name, extra)
+__all__.append("vdo_activate")
+
+_vdo_deactivate = BlockDev.vdo_deactivate
+@override(BlockDev.vdo_deactivate)
+def vdo_deactivate(name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_deactivate(name, extra)
+__all__.append("vdo_deactivate")
+
+_vdo_start = BlockDev.vdo_start
+@override(BlockDev.vdo_start)
+def vdo_start(name, rebuild=False, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_start(name, rebuild, extra)
+__all__.append("vdo_start")
+
+_vdo_stop = BlockDev.vdo_stop
+@override(BlockDev.vdo_stop)
+def vdo_stop(name, force=False, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_stop(name, force, extra)
+__all__.append("vdo_stop")
+
+_vdo_grow_logical = BlockDev.vdo_grow_logical
+@override(BlockDev.vdo_grow_logical)
+def vdo_grow_logical(name, size, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_grow_logical(name, size, extra)
+__all__.append("vdo_grow_logical")
+
+_vdo_grow_physical = BlockDev.vdo_grow_physical
+@override(BlockDev.vdo_grow_physical)
+def vdo_grow_physical(name, extra=None, **kwargs):
+    extra = _get_extra(extra, kwargs)
+    return _vdo_grow_physical(name, extra)
+__all__.append("vdo_grow_physical")
+
+
 ## defined in this overrides only!
 def plugin_specs_from_names(plugin_names):
     ret = []
@@ -818,7 +1005,7 @@ class ErrorProxy(object):
         """Let's make TAB-TAB in ipython work!"""
 
         if self._use_local:
-            items = set(dir(self._mod) + locals().keys())
+            items = set(dir(self._mod)) | set(locals().keys())
         else:
             items = set(dir(self._mod))
 
@@ -907,7 +1094,17 @@ __all__.append("MpathError")
 
 class SwapError(BlockDevError):
     pass
-__all__.append("SwapError")
+class SwapActivateError(SwapError):
+    pass
+class SwapOldError(SwapActivateError):
+    pass
+class SwapSuspendError(SwapActivateError):
+    pass
+class SwapUnknownError(SwapActivateError):
+    pass
+class SwapPagesizeError(SwapActivateError):
+    pass
+__all__.extend(("SwapError", "SwapActivateError", "SwapOldError", "SwapSuspendError", "SwapUnknownError", "SwapPagesizeError"))
 
 class KbdError(BlockDevError):
     pass
@@ -935,6 +1132,10 @@ class NVDIMMError(BlockDevError):
     pass
 __all__.append("NVDIMMError")
 
+class VDOError(BlockDevError):
+    pass
+__all__.append("VDOError")
+
 class BlockDevNotImplementedError(NotImplementedError, BlockDevError):
     pass
 __all__.append("BlockDevNotImplementedError")
@@ -942,6 +1143,11 @@ __all__.append("BlockDevNotImplementedError")
 not_implemented_rule = XRule(GLib.Error, re.compile(r".*The function '.*' called, but not implemented!"), None, BlockDevNotImplementedError)
 
 fs_nofs_rule = XRule(GLib.Error, None, 3, FSNoFSError)
+swap_activate_rule = XRule(GLib.Error, None, 1, SwapActivateError)
+swap_old_rule = XRule(GLib.Error, None, 3, SwapOldError)
+swap_suspend_rule = XRule(GLib.Error, None, 4, SwapSuspendError)
+swap_unknown_rule = XRule(GLib.Error, None, 5, SwapUnknownError)
+swap_pagesize_rule = XRule(GLib.Error, None, 6, SwapPagesizeError)
 
 btrfs = ErrorProxy("btrfs", BlockDev, [(GLib.Error, BtrfsError)], [not_implemented_rule])
 __all__.append("btrfs")
@@ -964,7 +1170,7 @@ __all__.append("md")
 mpath = ErrorProxy("mpath", BlockDev, [(GLib.Error, MpathError)], [not_implemented_rule])
 __all__.append("mpath")
 
-swap = ErrorProxy("swap", BlockDev, [(GLib.Error, SwapError)], [not_implemented_rule])
+swap = ErrorProxy("swap", BlockDev, [(GLib.Error, SwapError)], [not_implemented_rule, swap_activate_rule, swap_old_rule, swap_suspend_rule, swap_unknown_rule, swap_pagesize_rule])
 __all__.append("swap")
 
 kbd = ErrorProxy("kbd", BlockDev, [(GLib.Error, KbdError)], [not_implemented_rule])
@@ -984,3 +1190,6 @@ __all__.append("s390")
 
 utils = ErrorProxy("utils", BlockDev, [(GLib.Error, UtilsError)])
 __all__.append("utils")
+
+vdo = ErrorProxy("vdo", BlockDev, [(GLib.Error, VDOError)])
+__all__.append("vdo")
